# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.
"""Crop provision."""

import sys

if sys.version_info[:2] >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources

import logging
import os

import numpy as np
import pandas as pd
import rasterio

import inca
import inca.common.estat as estat
import inca.common.suts as suts
from inca.common.config_check import ConfigItem, check_csv, check_tsv, ConfigRaster, YEARLY
from inca.common.errors import Error, ConfigError
from inca.common.ecosystem import ECOTYPE, ECOTYPE_L1, ecotype_l1_cat, types_l1, ECO_ID, SHAPE_ID, \
    CROP, GRASSLAND
from inca.common.geoprocessing import statistics_byArea_byET, RasterType, number_blocks, block_window_generator
from inca.common.nuts import NUTS_ID

logger = logging.getLogger(__name__)

# some globals for faster adaptation
_CROPPROVISION = 'Crop Provision'
_OPERATION_MODE = 'operation-mode'
_LUT_CROPS = 'lut_crops'
_CROP_YIELD = 'crop_yield'
_PROXY_CROPTYPE = 'proxy_croptype'
_PROXY_CROPYIELD = 'proxy_cropyield'
_PROXY_PRODUCTIVITY = 'proxy_productivity'

# naming of calculation methods based on data source
MAT_FLOW   = 'material-flow'
AGRI_STATS = 'agriculture-statistics'
GEOSPATIAL = 'agriculture-yield-map'

CODE_MFA = 'code_mfa'
AACT_UV  = 'aact_uv01'
ECOCON   = 'ecocon'
CROPTYPE = 'code_croptype'

# units
EUR_TONNE = 'EUR_T'  # code for prices in euro per tonne in unit value table
EU_STD_HUMIDITY = 'PR_HU_EU'  # code for Harvested production in eu standard humidity (1000 t)
_UNIT = 'unit'
_THOUSAND_TONNE = 'THS_T'

_INDIC_ENV = 'indic_env'
_DOMESTIC_EXTRACTION = 'DE'


mfa_labels = {'MF11': 'MF1.1 crops (excluding fodder crops)',
              'MF111': 'MF1.1.1 cereals',
              'MF112': 'MF1.1.2 roots and tubers',
              'MF113': 'MF1.1.3 sugar crops',
              'MF114': 'MF1.1.4 pulses',
              'MF115': 'MF1.1.5 nuts',
              'MF116': 'MF1.1.6 oil-bearing crops',
              'MF117': 'MF1.1.7 vegetables',
              'MF118': 'MF1.1.8 fruits',
              'MF119': 'MF1.1.9 fibres',
              'MF11A': 'MF1.1.A other crops (excluding fodder crops) n.e.c.',
              'MF12':  'MF1.2 Crop residues (used) and fodder crops and grazed biomass',
              'MF121': 'MF1.2.1 Crop residues (used)',
              'MF1211': 'MF1.2.1.1 Straw',
              'MF1212': 'MF1.2.1.2 Other crop residues (sugar and fodder beet leaves and other)',
              'MF122': 'MF1.2.2 Fodder crops and grazed biomass',
              'MF1221': 'MF1.2.2.1 Fodder crops (including biomass harvest from grassland)',
              'MF1222': 'MF1.2.2.2 Grazed biomass'}

output_labels = {'SWHE': 'soft wheat',
                 'DWHE': 'durum wheat',
                 'BARL': 'barley',
                 'OATS': 'oats',
                 'MAIZ': 'maize',
                 'OCER': 'other cereals',
                 'RAPE': 'rape',
                 'SUNF': 'sunflower',
                 'MAIF': 'fodder maize',
                 'OFAR': 'other forage',
                 'PULS': 'protein crops',
                 'POTA': 'potatoes',
                 'SUGB': 'sugar beet'}


class CropRun(inca.Run):

    service_name = _CROPPROVISION

    def __init__(self, config):
        super().__init__(config)

        if config.get(_OPERATION_MODE) == MAT_FLOW:  # Currently no monetary evaluation for material flow
            self.monetary = False

        self.config_template.update({
            self.service_name: {
                _OPERATION_MODE: ConfigItem(default=MAT_FLOW),
                _LUT_CROPS: ConfigItem(check_csv, optional=True,
                                       required_columns=[CODE_MFA, 'label_mfa', ECOTYPE_L1]),
                _CROP_YIELD: ConfigItem(check_tsv),  # apro_cpsh1 table for AGRI_STATS, env_ac_mfa table for MAT_FLOW
                _PROXY_CROPTYPE: {YEARLY: ConfigRaster(optional=True, raster_type=RasterType.CATEGORICAL)},
                _PROXY_CROPYIELD: {YEARLY: ConfigRaster(optional=True, raster_type=RasterType.RELATIVE)},
                _PROXY_PRODUCTIVITY: {YEARLY: ConfigRaster(optional=True, raster_type=RasterType.RELATIVE)},
            }
        })
        if self.monetary:  # Inputs used only for monetary evaluation:
            logger.debug('Add config_template for monetary')
            self.config_template[self.service_name].update({
                AACT_UV: ConfigItem(check_tsv),
            })

        # if a developer wants to run the original KIP-INCA JRC method, some parameters are to be overruled
        # this method is not available from QGIS and requires a command line option
        if self.config.get('jrcmethod'):
            logger.warning('jrcmethod overwrite set.  Operation mode is %s; crop_yield should be in apro_cpsh1 format.',
                           AGRI_STATS)
            self.config_template[self.service_name].update({
                    ECOCON: ConfigItem(check_csv, required_columns=[NUTS_ID] + list(output_labels))})

            self.config[self.service_name][_OPERATION_MODE] = AGRI_STATS
            self.config['statistics_shape'] = 'NUTS-0'
            self.config['reporting_shape'] = 'NUTS-0'
            self.config['selected_regions'] = ['EU27']

        self.lut_crops = None
        self.make_output_filenames()

    def _start(self):

        # preprocess statistics input table
        production, crops = self.preprocess_yield_statistics()

        # filter on crops to account for operation mode: MFA, APRO, GEO # TODO
        if self.config.get('jrcmethod'):
            crops = self.lut_crops.dropna(subset=['code_CAPRI'])

        nyears = len(self.years)
        # loop over all processing years
        for year in self.years:
            logger.debug('year: %d', year)
            self.year = year

            # create empty frame for results
            ecosystems = self.lut_crops.ecosystem_type_L1.unique().to_list()
            self.result_tonne = pd.DataFrame(index=self.statistics_shape.index, columns=pd.Index([], name=CODE_MFA))
            self.result_tonne_new = pd.DataFrame(index=pd.MultiIndex.from_arrays(
                [list(np.repeat(self.statistics_shape.index.to_list(), len(ecosystems))), \
                        ecosystems * len (self.statistics_shape.index.to_list())], \
                        names=['NUTS_ID', ECOTYPE_L1]))
            self.result_eur_old = pd.DataFrame(index=self.statistics_shape.index, columns=pd.Index([], name=CODE_MFA))
            self.result_eur = pd.DataFrame(index=pd.MultiIndex.from_arrays(
                [list(np.repeat(self.statistics_shape.index.to_list(), len(ecosystems))), \
                 ecosystems * len(self.statistics_shape.index.to_list())], \
                names=['NUTS_ID', ECOTYPE_L1]))
            #no need for multi-index, since we keep crops from statistics in columns, only need for geospatial distribution
            self.result_tonne_croptype = pd.DataFrame(index=self.statistics_shape.index, columns=crops)

            # extract production statistics per crop and map to MFA
            self.crop_statistics(production, crops)

            # map MFA production statistics to ecosystem types, prepare for reporting
            self.supply_tonne = self.reporting_to_ET()

            # Disaggregate statistics using provided proxy/proxies or default via ecosystem extent map
            # get links to proxy/proxies, if available
            proxy_raster_crop_type = None
            proxy_raster_crop_yield = None
            proxy_raster_productivity = None
            if self.config[_CROPPROVISION][_PROXY_CROPYIELD][self.year]:
                proxy_raster_crop_yield = self.config[self.service_name][_PROXY_CROPYIELD][self.year]
            if self.config[_CROPPROVISION][_PROXY_CROPTYPE][self.year]:
                proxy_raster_crop_type = self.config[self.service_name][_PROXY_CROPTYPE][self.year]
            if self.config[_CROPPROVISION][_PROXY_PRODUCTIVITY][self.year]:
                proxy_raster_productivity = self.config[self.service_name][_PROXY_PRODUCTIVITY][self.year]
            # generate map based on statistics
            self.crop_provision_map_generation(proxy_raster_crop_type, proxy_raster_crop_yield, proxy_raster_productivity,
                                               add_progress=lambda p: self.add_progress(0.5 * p / nyears))
            self.cut_to_reporting_regions(self.flow_map_physical[year])

            # Statistics extraction for reporting regions
            self.reporting_area_statistics(add_progress=lambda p: self.add_progress(0.5 * p / nyears))

            # Export of reporting area statistics to SUT Exel file and CSV statistic files
            self.reporting()

        return

    def make_output_filenames(self):
        """Create standard output file names."""
        crop_provision = _CROPPROVISION.lower().replace(' ', '-')  # lower case, without spaces
        self.flow_map_physical = {year: os.path.join(self.maps,
                                                     f'{crop_provision}_map_use_1000-tonne_{year}.tif')
                                  for year in self.years}
        self.flow_map_monetary_real = {year: os.path.join(self.maps,
                                                          f'{crop_provision}_map_use_M-EURO-real_{year}.tif')
                                       for year in self.years}
        self.flow_map_monetary_nominal = {year: os.path.join(self.maps,
                                                             f'{crop_provision}_map_use_M-EURO-current_{year}.tif')
                                          for year in self.years}
        self.use_physical = {year: os.path.join(self.statistic_dir, f'{crop_provision}_table_use_1000-tonne_{year}.csv')
                             for year in self.years}
        self.use_monetary = {year: os.path.join(self.statistic_dir,
                                                f'{crop_provision}_table_use_M-EURO-real_{year}.csv')
                             for year in self.years}
        self.use_monetary_nominal = {year: os.path.join(self.statistic_dir,
                                                        f'{crop_provision}_table_use_M-EURO-current_{year}.csv')
                                     for year in self.years}
        self.supply_physical = {year: os.path.join(self.statistic_dir,
                                                   f'{crop_provision}_table_supply_1000-tonne_{year}.csv')
                                for year in self.years}
        self.supply_monetary = {year: os.path.join(self.statistic_dir,
                                                   f'{crop_provision}_table_supply_M-EURO-real_{year}.csv')
                                for year in self.years}
        self.supply_monetary_nominal = {year: os.path.join(self.statistic_dir,
                                                           f'{crop_provision}_table_supply_M-EURO-current_{year}.csv')
                                        for year in self.years}
        self.sut_physical = {year: os.path.join(self.suts,
                                                f'{crop_provision}_report_SUT-physical_1000-tonne_{year}.xlsx')
                             for year in self.years}
        self.sut_monetary_real = {year: os.path.join(self.suts,
                                                     f'{crop_provision}_report_SUT-monetary_M-EURO-real_{year}.xlsx')
                                  for year in self.years}
        self.sut_monetary_nominal = {year: os.path.join(
            self.suts, f'{crop_provision}_report_SUT-monetary_M-EURO-current_{year}.xlsx')
            for year in self.years}

    def preprocess_yield_statistics(self):
        """Retrieve crop statistics and map to supply table format (MFA based)."""
        # read crop production:
        production = estat.read_tsv(self.config[self.service_name][_CROP_YIELD])

        # Check _CROP_YIELD has the right format:
        # For MAT_FLOW, crop yield table should have the env_ac_mfa.tsv format, for AGRI_STATS, we should have the
        # apro_cpsh1.tsv format.
        if self.config[_CROPPROVISION][_OPERATION_MODE] == MAT_FLOW:
            self.code = CODE_MFA
            expected_index_levels = ('indic_env', 'material', 'unit')
            standard_file = 'env_ac_mfa.tsv'
        elif self.config[_CROPPROVISION][_OPERATION_MODE] == AGRI_STATS:
            self.code = 'code_apro'
            expected_index_levels = ('crops', 'strucpro')
            standard_file = 'apro_cpsh1.tsv'
        elif self.config[_CROPPROVISION][_OPERATION_MODE] in [GEOSPATIAL]:
            raise ConfigError('operation-mode geospatial is not yet supported', [_CROPPROVISION, _OPERATION_MODE])
        else:
            raise ConfigError(f'operation-mode was set to "{self.config[_CROPPROVISION][_OPERATION_MODE]}", '
                              f'should be set to "{MAT_FLOW}" or "{AGRI_STATS}".',
                              [_CROPPROVISION, _OPERATION_MODE])
        # Check index has the expected levels
        if any(level not in production.index.names for level in expected_index_levels):
            raise ConfigError(f'Crop yield file "{self.config[self.service_name][_CROP_YIELD]}" does not have the '
                              f'correct format.  For operation mode {self.config[self.service_name][_OPERATION_MODE]}, '
                              f'the file should have the same structure as Eurostat standard dataset {standard_file}.',
                              [self.service_name, _CROP_YIELD])

        # read lookup crop mapping table from MFA to (optional) APRO and or GEO
        # If no file provided by user, use default LUT:
        ref = importlib_resources.files(inca.data).joinpath('cropprovision_lut.csv')
        file = self.config[self.service_name][_LUT_CROPS] if self.config[self.service_name][_LUT_CROPS] else ref
        with open(file, 'r') as f:
            lut_crops = pd.read_csv(f, comment='#', dtype={'code_val': str, ECOTYPE_L1: ecotype_l1_cat})

        if self.code not in lut_crops:
            raise ConfigError(f'Missing column in file "{self.config[self.service_name][_LUT_CROPS]}".  For method '
                              f'"{_OPERATION_MODE}" method, the crop lookup table must have a column "{self.code}".',
                              [_CROPPROVISION, _LUT_CROPS])
        self.lut_crops = lut_crops.set_index(self.code)

        # clean production statistics table
        assert (self.code in ('code_apro', CODE_MFA)), \
            'Crop provision service supports only material flows "env_ac_mfa" or agriculture statistics "apro_cpsh1"'
        if self.code == 'code_apro':
            production = production.xs(EU_STD_HUMIDITY, level='strucpro')
        elif self.code == CODE_MFA:
            # cross-sect dataframe Domestic Extraction, keep croptype and geo
            production = production.xs((_DOMESTIC_EXTRACTION, _THOUSAND_TONNE), level=(_INDIC_ENV, _UNIT))
            self.lut_crops[CODE_MFA] = self.lut_crops.index

        # read unit values at basic prices:
        if self.monetary:
            self.unit_value = estat.read_tsv(self.config[self.service_name][AACT_UV])
            self.unit_value = self.unit_value.xs(EUR_TONNE, level='unit')
            # Manually replace 0. by pd.NA, as method .replace(0., pd.NA) throws RecursionError on some pandas versions.
            for col in self.unit_value.columns:
                self.unit_value.loc[self.unit_value[col] == 0., col] = pd.NA

        # subset based on reporting areas
        production = production.reindex(self.statistics_shape.index, level=1, fill_value=np.nan)
        crops = production.index.unique(level=0)  # apro: level 0 == 'crops', mfa: level 0 == 'material'
        if self.monetary:
            self.unit_value = self.unit_value.reindex(self.statistics_shape.index, level=1, fill_value=np.nan)

        return production, crops

    def crop_statistics(self, production, crops):
        """Map crop statistics to supply table format (MFA based)."""
        # read ecological contribution (only in KIP-INCA JRC-method mode)
        if self.config.get('jrcmethod'):
            ecocon = pd.read_csv(self.config[self.service_name][ECOCON], comment='#').set_index(NUTS_ID)
            crops = crops.index
            logger.debug('ecocon:\n%s', ecocon)

        # extract statistics per crop
        for crop in crops:
            logger.debug('extract statistics for crop %s', crop)
            #TEST
            #if (crop == 'S0000') or (crop == 'R0000') or (crop == 'G3000'):
            #    a=1

            if (crop not in self.lut_crops.index) or (np.array(pd.isnull(self.lut_crops.loc[crop].code_mfa)).any()):
                logger.warning('Skipping crop %s: not in lookup table or not mapped to Material-flow', crop)
                continue

            if self.monetary and \
                    (not pd.isna(self.lut_crops.loc[crop].code_val)) and \
                    (str(self.lut_crops.loc[crop].code_val).zfill(5) in self.unit_value.index.get_level_values(0).to_list()):
                price = mean_price(self.unit_value.loc[self.lut_crops.loc[crop].code_val.zfill(5)], self.year,
                                   self.lut_crops.loc[crop].label_val)  # [eur/tonne]
            else:
                # TODO need to find source for prices in euro-tonnes for material flow
                price = np.nan
            tmp = pd.DataFrame(index=estat.eu_members(2020))

            # KIP-INCA JRCmethod takes average over 3 years, and uses ecosystem contribution 'emergy' approach
            if self.config.get('jrcmethod'):
                # calculate mean over 3 years
                for y in (self.year - 1, self.year, self.year + 1):
                    if y in production.columns:
                        tmp[y] = production.loc[crop, y]
                    else:
                        logger.warning('No production value for crop %s in year %d',
                                       self.lut_crops.loc[crop].code_prod, y)
                prod = tmp.mean(axis=1)  # [1000 tonnes]
                crop_tonne = prod * ecocon[self.lut_crops.loc[crop].code_CAPRI]
            else:
                crop_tonne = production.loc[crop, self.year]  # in thousand tonnes

            # express monetary valuation in millions euro
            crop_eur = crop_tonne * price / 1000

            # append results to dataframes (one for physical and one for monetary)
            colname = self.lut_crops.loc[crop].code_mfa   # use code, transform later to label
            et_code = self.lut_crops.loc[crop].ecosystem_type_L1
            if not isinstance(colname, str):
                colname = colname.unique()[0]
            if isinstance(et_code, pd.core.series.Series):
                if len(et_code.unique().to_list()) > 1:
                    logger.warning('Multiple ET codes found for crop {}: {}'.format(colname, et_code.unique().to_list()))
                et_code =et_code.unique()[0]

            if colname in self.result_tonne.columns:  # if multiple crops go in one CAPRI category
                self.result_tonne[colname] = self.result_tonne[colname].add(crop_tonne, fill_value=0.)
                self.result_tonne_croptype[crop] = self.result_tonne_croptype[crop].add(crop_tonne, fill_value=0.)
                for geo in crop_tonne.index.to_list():
                    if np.isnan(self.result_tonne_new.loc[(geo,et_code),colname]): self.result_tonne_new.loc[(geo,et_code),colname]=0
                    if np.isnan(crop_tonne[geo]): continue
                    self.result_tonne_new.loc[(geo,et_code),colname] += crop_tonne[geo]
                    self.result_eur.loc[(geo, et_code), colname] += crop_eur[geo]
                #self.result_eur[colname] = self.result_eur[colname].add(crop_eur, fill_value=0.)
            else:
                self.result_tonne[colname] = crop_tonne
                self.result_tonne_croptype[crop] = crop_tonne  # keep original to distribute if croptype map
                for geo in crop_tonne.index.to_list():
                    self.result_tonne_new.loc[(geo,et_code),colname] = crop_tonne[geo]
                    self.result_eur.loc[(geo, et_code), colname] = crop_eur[geo]
                #self.result_eur[colname] = crop_eur

        # "real" value in terms of 2000 inflation index
        if self.monetary:
            self.result_eur_real = self.result_eur.multiply(self.deflator[2000] / self.deflator[self.year],
                                                            axis='index')
            logger.debug('deflator values:\n%s', self.deflator[2000] / self.deflator[self.year])

        # remove subtotal row, avoid double count
        mfa_key_subtotals = ['MF11', 'MF12', 'MF121', 'MF122']
        for mfa in mfa_key_subtotals:
            if mfa in self.result_tonne.columns:
                self.result_tonne.drop(mfa, axis=1, inplace=True)
                self.result_tonne_new.drop(mfa, axis=1, inplace=True)
                self.result_tonne_croptype.drop(mfa, axis=1, inplace=True)
            if self.monetary:
                if mfa in self.result_eur.columns:
                    self.result_eur.drop(mfa, axis=1, inplace=True)
                if mfa in self.result_eur_real.columns:
                    self.result_eur_real.drop(mfa, axis=1, inplace=True)

        # add missing MFA types
        mfa_keys = list(mfa_labels.keys())
        mfa_keys = [x for x in mfa_keys if x not in mfa_key_subtotals]
        for mfa in mfa_keys:
            if mfa not in self.result_tonne.columns:
                self.result_tonne[mfa] = 0
                self.result_tonne_new.loc[:,mfa] = 0
            if self.monetary:
                if mfa not in self.result_eur.columns:
                    self.result_eur[mfa] = 0
                if mfa not in self.result_eur_real.columns:
                    self.result_eur_real[mfa] = 0

        if self.monetary:
            logger.debug('result_eur:\n%s', self.result_eur.sum(axis=1))
        logger.debug('result_tonne:\n%s', self.result_tonne_new.sum(axis=1))

        return

    def crop_provision_map_generation(self, proxy_raster_file_type, proxy_raster_file_yield, proxy_raster_productivity, add_progress):
        """Generate the crop provision maps for the statistical regions."""
        logger.info('start generation of crop provision map for year %s', self.year)

        # Create dataframe indexed by NUTS_ID and ecosystem type:
        df_data = self.supply_tonne.droplevel(CODE_MFA)
        df_data = df_data.groupby([NUTS_ID, ECOTYPE_L1]).sum()
        # Make sure index contains all ecotypes, and rename to ECOTYPE instead of ECOTYPE_L1:
        idx = pd.MultiIndex.from_product([df_data.index.unique(NUTS_ID), types_l1[ECOTYPE]])
        df_data = df_data.reindex(idx, fill_value=0.)

        if not proxy_raster_file_yield:
            #no crop yield geospatial information available
            progress_weight_physical = 0.4 if self.monetary else 1.
            if proxy_raster_file_type is not None:
                logger.info('geospatial distribution through crop type map.')
                #prepare dataframe from original croptype table as it contains more details (in agriculture statistics)
                df_data = self.result_tonne_croptype.dropna(axis=1)
                #replace croplabels with croptype raster values
                lut = self.lut_crops[[CROPTYPE,ECOTYPE_L1]].copy().dropna()
                lut = lut[lut.index.notna()]
                #check duplicates in croptype codes mapping to statistics (esp. using MFA statistics)
                lut_duplicates_d = lut[lut.index.duplicated()].drop(ECOTYPE_L1, axis=1)
                lut_duplicates_d = lut_duplicates_d.groupby(level=0).apply(lambda x: x.to_dict('r')).to_dict()
                lut_duplicates = dict()
                for k,v in lut_duplicates_d.items():
                    for dic in v:
                        lut_duplicates[int(dic[CROPTYPE])] = k
                #group all croplabels missing in croptype raster
                lut_mapping = dict(zip(lut.index, lut[CROPTYPE].astype(np.uint)))
                lut_missing = [x for x in df_data.columns.to_list() if x not in lut.index.to_list()]
                for crop in lut_missing:
                    if isinstance(self.lut_crops.loc[crop],pd.Series):
                        lut_mapping[crop] = self.lut_crops.loc[crop].ecosystem_type_L1
                    else:
                        lut_mapping[crop] = self.lut_crops.loc[crop].drop_duplicates().ecosystem_type_L1.values[0]
                #set replacement values for raster duplicates
                for k,v in lut_duplicates.items():
                    lut_duplicates[k] = lut_mapping[v]
                #convert lut_mapping to string
                new_lut_mapping = dict()
                #check if mapping table is correct (no nans left)
                if any(np.isnan(val) for val in lut_mapping.values()):
                    logger.exception('Some crop types not mapped to an ecosystem type, correct lookup table configuration file')
                #prepare dataframe for distribution of statistics
                for i in list(lut_mapping.keys()):
                    new_lut_mapping[i] = str(lut_mapping[i])
                df_data.columns.name = 'crops'
                df_data = df_data.stack().to_frame(name='physical').rename(index=new_lut_mapping)
                df_data = df_data.groupby(['NUTS_ID','crops']).sum()
                df_data.index.set_names(['NUTS_ID', ECOTYPE], inplace=True)

                # detail ecosystem type map with crop type information
                proxy_raster_file_type_ET = self.imprint_croptype_map(proxy_raster_file_type, df_data, lut, lut_duplicates,
                                    add_progress=lambda p: add_progress(progress_weight_physical * 0.5 * p))

                # finally we can create the geospatial map
                if not proxy_raster_productivity:
                    logger.info('no productivity proxy provided, distribute yield equally per crop type.')
                    self.accord.spatial_disaggregation_byArea_byET(
                        proxy_raster_file_type_ET, df_data['physical'],
                        self.statistics_raster, self.statistics_shape[SHAPE_ID],
                        proxy_raster_file_type_ET, dict(zip(map(str,df_data.index.levels[1].to_list()),\
                                                            [eval(i) for i in df_data.index.levels[1].to_list()])),
                        self.flow_map_physical[self.year],
                        add_progress=lambda p: add_progress(progress_weight_physical * p),
                        unit_info='1000 tonnes', processing_info='Spatial disaggregation through crop type map.')
                    '''if self.monetary:
                        self.accord.spatial_disaggregation_byArea_byET(
                            proxy_raster_file_type_ET, df_data['monetary'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID],
                            proxy_raster_file_type_ET, dict(zip(map(str, df_data.index.levels[1].to_list()), \
                                                                [eval(i) for i in df_data.index.levels[1].to_list()])),
                            self.flow_map_monetary_nominal[self.year],
                            add_progress=lambda p: add_progress(progress_weight_physical * p),
                            unit_info='Million EUR', processing_info='Spatial disaggregation of use through crop type map.')
                        self.accord.spatial_disaggregation_byArea_byET(
                            proxy_raster_file_type_ET, df_data['monetary_real'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID],
                            proxy_raster_file_type_ET, dict(zip(map(str, df_data.index.levels[1].to_list()), \
                                                                [eval(i) for i in df_data.index.levels[1].to_list()])),
                            self.flow_map_monetary_real[self.year],
                            add_progress=lambda p: add_progress(progress_weight_physical * p),
                            unit_info='Million EUR (Reference year 2000 value)',
                            processing_info='Spatial disaggregation of use through crop type map.')
                    '''
                else:
                    self.accord.spatial_disaggregation_byArea_byET(
                        proxy_raster_productivity, df_data['physical'],
                        self.statistics_raster, self.statistics_shape[SHAPE_ID],
                        proxy_raster_file_type_ET, dict(zip(map(str, df_data.index.levels[1].to_list()),
                                                            [eval(i) for i in df_data.index.levels[1].to_list()])),
                        self.flow_map_physical[self.year],
                        add_progress=lambda p: add_progress(progress_weight_physical * p),
                        unit_info='1000 tonnes', processing_info='Spatial disaggregation through crop type and GPP maps.')
                    '''if self.monetary:
                        self.accord.spatial_disaggregation_byArea_byET(
                            proxy_raster_productivity, df_data['monetary'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID],
                            proxy_raster_file_type_ET, dict(zip(map(str, df_data.index.levels[1].to_list()), \
                                                                [eval(i) for i in df_data.index.levels[1].to_list()])),
                            self.flow_map_monetary_nominal[self.year],
                            add_progress=lambda p: add_progress(progress_weight_physical * p),
                            unit_info='Million EUR', processing_info='Spatial disaggregation of use through crop type and GPP maps.')
                        self.accord.spatial_disaggregation_byArea_byET(
                            proxy_raster_productivity, df_data['monetary_real'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID],
                            proxy_raster_file_type_ET, dict(zip(map(str, df_data.index.levels[1].to_list()), \
                                                                [eval(i) for i in df_data.index.levels[1].to_list()])),
                            self.flow_map_monetary_real[self.year],
                            add_progress=lambda p: add_progress(progress_weight_physical * p),
                            unit_info='Million EUR (Reference year 2000 value)',
                            processing_info='Crop provision supply map for year '
                                            f'{self.year} through geospatial proxies crop type and GPP maps.')
                    '''
            else:
                logger.warning('no proxy map provided for crop-types and/or crop-yield, '
                               'use default ecosystem extent map.')
                # disaggregate by ecosystem type map
                if not proxy_raster_productivity:
                    self.accord.spatial_disaggregation_byArea_byET(
                        self.get_ecosystem_raster(self.year, level=1), df_data['physical'],
                        self.statistics_raster, self.statistics_shape[SHAPE_ID].to_dict(),
                        self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID].to_dict(),
                        self.flow_map_physical[self.year],
                        add_progress=add_progress, unit_info='1000 tonnes',
                        processing_info='Crop provision supply map for year '
                        f'{self.year} using the ecosystem extent map as proxy')
                    '''if self.monetary:
                        self.accord.spatial_disaggregation_byArea_byET(
                            self.get_ecosystem_raster(self.year, level=1), df_data['monetary'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID].to_dict(),
                            self.get_ecosystem_raster(self.year, level=1),
                            types_l1.set_index(ECOTYPE)[ECO_ID].to_dict(),
                            self.flow_map_monetary_nominal[self.year],
                            add_progress=add_progress, unit_info='Million EUR',
                            processing_info='Crop provision supply map for year '
                                            f'{self.year} using the ecosystem extent map as proxy')
                        self.accord.spatial_disaggregation_byArea_byET(
                            self.get_ecosystem_raster(self.year, level=1), df_data['monetary_real'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID].to_dict(),
                            self.get_ecosystem_raster(self.year, level=1),
                            types_l1.set_index(ECOTYPE)[ECO_ID].to_dict(),
                            self.flow_map_monetary_real[self.year],
                            add_progress=add_progress, unit_info='Million EUR (Reference year 2000 value)',
                            processing_info='Crop provision supply map for year '
                                            f'{self.year} using the ecosystem extent map as proxy')
                    '''
                else:
                    self.accord.spatial_disaggregation_byArea_byET(
                        proxy_raster_productivity, df_data['physical'],
                        self.statistics_raster, self.statistics_shape[SHAPE_ID].to_dict(),
                        self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID].to_dict(),
                        self.flow_map_physical[self.year],
                        add_progress=add_progress, unit_info='1000 tonnes',
                        processing_info='Spatial disaggregation of use the ecosystem extent map and GPP as proxy')
                    '''if self.monetary:
                        self.accord.spatial_disaggregation_byArea_byET(
                            proxy_raster_productivity, df_data['monetary'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID].to_dict(),
                            self.get_ecosystem_raster(self.year, level=1),
                            types_l1.set_index(ECOTYPE)[ECO_ID].to_dict(),
                            self.flow_map_monetary_nominal[self.year],
                            add_progress=add_progress, unit_info='Million EUR',
                            processing_info='Crop provision supply map for year '
                                            f'{self.year} using the ecosystem extent map and GPP as proxy')
                        self.accord.spatial_disaggregation_byArea_byET(
                            proxy_raster_productivity, df_data['monetary_real'],
                            self.statistics_raster, self.statistics_shape[SHAPE_ID].to_dict(),
                            self.get_ecosystem_raster(self.year, level=1),
                            types_l1.set_index(ECOTYPE)[ECO_ID].to_dict(),
                            self.flow_map_monetary_real[self.year],
                            add_progress=add_progress, unit_info='Million EUR (Reference year 2000 value)',
                            processing_info='Crop provision supply map for year '
                                        f'{self.year} using the ecosystem extent map and GPP as proxy')
                    '''

        else:  # have proxy_raster_file_yield:
            if proxy_raster_file_type is not None:
                logger.warning('geospatial distribution through crop type map not yet supported, use yield map only')
            else:
                logger.warning('no proxy map provided for crop-types, use yield map only')

            # If we have monetary evaluation, ~60% of time is spent on that, and 40% on physical part.  Without monetary
            # evaluation, 100% of time is spent on the physical part.
            progress_weight_physical = 0.4 if self.monetary else 1.
            proxy_sums = self.accord.spatial_disaggregation_byArea_byET(
                proxy_raster_file_yield, df_data['physical'],
                self.statistics_raster, self.statistics_shape[SHAPE_ID],
                self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID],
                self.flow_map_physical[self.year],
                add_progress=lambda p: add_progress(progress_weight_physical * p),
                unit_info='1000 tonnes', processing_info='Spatial disaggreation of use.')
            if self.monetary:
                self.accord.spatial_disaggregation_byArea_byET(
                    proxy_raster_file_yield, df_data['monetary'],
                    self.statistics_raster, self.statistics_shape[SHAPE_ID],
                    self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID],
                    self.flow_map_monetary_nominal[self.year],
                    add_progress=lambda p: self.add_progress(0.3 * p),
                    proxy_sums=proxy_sums,
                    unit_info='Million EUR', processing_info='Spatial disaggregation of use.')
                self.accord.spatial_disaggregation_byArea_byET(
                    proxy_raster_file_yield, df_data['monetary_real'],
                    self.statistics_raster, self.statistics_shape[SHAPE_ID],
                    self.get_ecosystem_raster(self.year, level=1), types_l1.set_index(ECOTYPE)[ECO_ID],
                    self.flow_map_monetary_real[self.year],
                    add_progress=lambda p: self.add_progress(0.3 * p),
                    proxy_sums=proxy_sums,
                    unit_info='Million EUR (Reference year 2000 value)',
                    processing_info='Spatial disaggregation of use.')
        return

    def imprint_croptype_map(self, proxy_raster_file_type, df_data, lut, lut_duplicates, add_progress, BLOCKSIZE=(2048,2048)):
        """Detail ecosystem type map with crop type map, through imprinting."""
        logger.info("Imprint crop types from {} into ecosystem type L1 map".format(proxy_raster_file_type))
        progress_remain = 1.
        with rasterio.open(self.get_ecosystem_raster(self.year, level=1), 'r') as src_et, \
                rasterio.open(proxy_raster_file_type, 'r') as src_crop:
            dst_profile = src_crop.profile
            dst_profile.update(dtype=rasterio.uint16, nodata=0)
            # number of to process blocks
            nblocks = number_blocks(dst_profile, BLOCKSIZE)

            # init the output raster file
            proxy_raster_file_type_ET = os.path.join(os.path.split(proxy_raster_file_type)[0],
                                os.path.splitext(os.path.split(proxy_raster_file_type)[1])[0] + '_ET.tif')
            with rasterio.open(proxy_raster_file_type_ET, 'w', **dst_profile) as dst:
                # add metadata if needed
                # self.metadata.update_dataset_tags(dst, '', '', self.get_ecosystem_raster(self.year, level=1),
                #                                  proxy_raster_file_type)

                countRemovals = 0
                countAdditions = 0
                countTotal = 0
                # now iterate over the blocks
                for _, window in block_window_generator(BLOCKSIZE, dst_profile['height'], dst_profile['width']):
                    # read data
                    aEcosystems = src_et.read(1, window=window, masked=True)
                    aCrops = src_crop.read(1, window=window, masked=True)

                    countTotal += aEcosystems[aEcosystems <= lut.ecosystem_type_L1.astype(np.uint16).unique().max()].count()

                    aData = np.zeros(aCrops.shape, dtype=np.uint16)
                    for crop in [int(i) for i in df_data.index.levels[1].to_list()]:
                        aMask = (aCrops == crop)
                        # overwrite if duplicated raster values for any statistic
                        if crop in lut_duplicates.keys():
                            crop = lut_duplicates[crop]
                        aData[aMask] = crop

                    # remove crop fields if not matching to ecosystem type
                    aMaskET = np.zeros(aCrops.shape, dtype=np.uint16)
                    for ET in self.lut_crops.ecosystem_type_L1.astype(np.uint16).unique():
                        aMask = (aEcosystems == ET)
                        aMaskET[aMask] = ET
                    aMask = ((aData != 0) & (aMaskET == 0))
                    countRemovals += np.count_nonzero(aMask)
                    aData[aMask] = 0

                    # assign missing crop types to ecosystem type
                    aMask = ((aData == 0) & (aMaskET != 0))
                    countAdditions += np.count_nonzero(aMask)
                    aData[aMask] = aMaskET[aMask]

                    # write to disk
                    dst.write(aData, 1, window=window)
                    add_progress(progress_remain / nblocks)

        logger.warning('{} pixels ({:.1f}%) from crop type map removed, inconsistent with ET map'.format(countRemovals, (
                    countRemovals / countTotal * 100)))
        logger.warning('{} pixels ({:.1f}%) from ET added, missing in crop type map'.format(countAdditions, (
                    countAdditions / countTotal * 100)))

        return proxy_raster_file_type_ET

    def reporting(self):
        """Write statistics csv files and SUT spreadsheets."""
        logger.info('statistics export')
        logger.info('* CSV statistic files')

        # in CSV files we use the shortnames for ecosystem types
        ecotype_short_name = types_l1.set_index(ECOTYPE)['short_name']
        # merge with short names, and keep only physical and short_name columns
        result_tonne_csv = self.supply_tonne.merge(ecotype_short_name,
                                                   left_on=ECOTYPE_L1, right_index=True).sort_index()[['physical',
                                                                                                       'short_name']]

        # replace MFA codes by longer MFA labels
        result_tonne_csv.rename(index=mfa_labels, inplace=True)
        result_tonne_csv.index.rename('MFA', level=CODE_MFA, inplace=True)

        # replace column labels
        result_tonne_csv.columns = ['supply', 'ecosystem type']

        result_tonne_csv.to_csv(self.supply_physical[self.year])
        '''
        self.reporting_eur.to_csv(self.use_monetary[self.year], index_label='NUTS_ID')
        self.reporting_eur_real.to_csv(self.supply_monetary[self.year], index_label='NUTS_ID')
        '''

        logger.info('* SUT Excel files')
        # supply/use tables: all use is intermediate consumption by industries, all supply is from cropland
        suts.write_sut(self.sut_physical[self.year],
                       self.reporting_tonne.unstack(),
                       self.reporting_tonne.unstack().sum(axis=1).rename(suts.IND),
                       self.service_name, self.year, '1000 tonnes')
        '''if self.monetary:
            suts.write_sut(self.sut_monetary_nominal[self.year],
                           self.reporting_eur.unstack(),
                           self.reporting_eur.unstack().sum(axis=1).rename(suts.IND),
                           self.service_name, self.year, 'Million EUR (2000 value)')
            suts.write_sut(self.sut_monetary_real[self.year],
                           self.reporting_eur_real.unstack(),
                           self.reporting_eur_real.unstack().sum(axis=1).rename(suts.IND),
                           self.service_name, self.year, 'Million EUR (nominal)')
        '''

    def reporting_to_ET(self):
        """Map table crops to ecosystem types."""
        # dataframe with physical values
        df = self.result_tonne_new.stack(dropna=False).reset_index(level=1).rename(columns={0: 'physical'})
        df.index.names = ['NUTS_ID','code_mfa']
        df.set_index(ECOTYPE_L1,append=True, inplace=True)
        if self.monetary:
            # dataframe with monetary values
            df2 = self.result_eur.stack(dropna=False).reset_index(level=1).rename(columns={0: 'monetary'})
            df2.index.names = ['NUTS_ID', 'code_mfa']
            df2.set_index(ECOTYPE_L1,append=True, inplace=True)
            # dataframe with monetary real values
            df3 = self.result_eur_real.stack(dropna=False).reset_index(level=1).rename(columns={0: 'monetary_real'})
            df3.index.names = ['NUTS_ID', 'code_mfa']
            df3.set_index(ECOTYPE_L1, append=True, inplace=True)
            # join into one single dataframe
            df = df.join(df2, on=['NUTS_ID', 'code_mfa', 'ecosystem_type_L1'])
            df = df.join(df3, on=['NUTS_ID', 'code_mfa', 'ecosystem_type_L1'])

        return df

        # Move CODE_MFA to index:
        df = self.result_tonne.stack().sort_index().rename('physical').to_frame()
        if self.monetary:
            df['monetary'] = self.result_eur.stack()
            df['monetary_real'] = self.result_eur_real.stack()

        # Create a mapping from CODE_MFA to ecosystem type:
        mfa_ecotype_mapping = self.lut_crops[[ECOTYPE_L1, CODE_MFA]].dropna().drop_duplicates(subset=CODE_MFA).dropna()
        mfa_ecotype_mapping.set_index(CODE_MFA, inplace=True)

        types_l1_idx = types_l1.set_index('name')
        cropland_id = types_l1_idx.loc[CROP, ECOTYPE]
        grassland_id = types_l1_idx.loc[GRASSLAND, ECOTYPE]

        # assign missing MFA to ET (MFA11x to cropland, MFA12x to grassland)
        for mfa in self.result_tonne.columns:
            if mfa not in mfa_ecotype_mapping.index:
                if mfa.startswith('MF11'):
                    mfa_ecotype_mapping.loc[mfa] = cropland_id
                elif mfa.startswith('MF12'):
                    mfa_ecotype_mapping.loc[mfa] = grassland_id
                else:
                    raise Error(f"Don't know which ecosystem type to assign to crop with MFA code {mfa}.")

        # add ecosystem type column to dataframe by joining with mfa_mapping
        return df.join(mfa_ecotype_mapping)

    def reporting_area_statistics(self, add_progress):
        """Extract statistics for the reporting regions."""
        logger.info('extract statistics for reporting regions')

        # If we have monetary evaluation, ~60% of time is spent on that, and 40% on physical part.  Without monetary
        # evaluation, 100% of time is spent on the physical part.
        progress_weight_physical = 0.4 if self.monetary else 1.

        ecosystem_raster = self.get_ecosystem_raster(self.year, level=1,
                                                     add_progress=lambda p: add_progress(
                                                         progress_weight_physical * 0.5 * p))

        self.reporting_tonne = statistics_byArea_byET(self.flow_map_physical[self.year],
                                                      self.reporting_raster, self.reporting_shape[SHAPE_ID],
                                                      ecosystem_raster,
                                                      types_l1.set_index('name')[ECO_ID],
                                                      lambda p: add_progress(progress_weight_physical * 0.5 * p))['sum']

        '''if self.monetary:
            self.reporting_eur = statistics_byArea_byET(self.flow_map_monetary_nominal[self.year],
                                                        self.reporting_raster, self.reporting_shape[SHAPE_ID],
                                                        ecosystem_raster,
                                                        types_l1.set_index('name')[ECO_ID],
                                                        lambda p: add_progress(0.3 * p))['sum']
            self.reporting_eur_real = statistics_byArea_byET(self.flow_map_monetary_real[self.year],
                                                             self.reporting_raster, self.reporting_shape[SHAPE_ID],
                                                             ecosystem_raster,
                                                             types_l1.set_index('name')[ECO_ID],
                                                             lambda p: add_progress(0.3 * p))['sum']
        '''

def set_cmdline_args(parser):
    """Implement additional command line arguments.

    :param parser: subparser object for CropProvision case in the __main__
    :return:
    """
    parser.add_argument('--jrcmethod', action='store_true',
                        help='To run the crop provision in original KIP-INCA JRC method (emergy function)')

def mean_price(unit_value, year, cropname):
    """Calculate the average of the price for (year-1, year, year+1) from the unit_value table for a crop.

    Gap-filling:
    - if no prices available for those years, use the price from the nearest year for the same member state.
    - if no price for that member state at all, use the EU average for that year.
    """
    assert (not (unit_value == 0.).any(axis=None))  # prices of 0. must be replaced by NA values here.
    tmp = pd.DataFrame(index=unit_value.index)
    for y in (year - 1, year, year + 1):
        if y in unit_value.columns:
            tmp[y] = unit_value[y]
        else:
            logger.warning('No data for year %d and crop %s in unit_value table.', y, cropname)
    result = tmp.mean(axis=1)
    eu_average = estat.filter_eu27(result).mean()
    # countries where 3-year average is 0 or missing:
    for member_state in result.index[result.isna()]:
        # get all available non-zero prices for that member state:
        ms_prices = unit_value.loc[member_state].dropna()
        if ms_prices.empty:  # no actual prices for this member state
            result.loc[member_state] = eu_average
            logger.debug('Use eu average price for %d %s %s: %f', year, member_state, cropname, eu_average)
        else:  # find the price from the nearest year
            idx_nearest_year = np.argmin(np.abs(ms_prices.index - year))
            result.loc[member_state] = ms_prices.iloc[idx_nearest_year]
            logger.debug('Use price from nearest year (%d) for %d %s %s: %f',
                         ms_prices.index[idx_nearest_year],
                         year, member_state, cropname, result.loc[member_state])

    return result
